# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Plugin resources."""
from __future__ import annotations

import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_JOBS_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    JobGetQueryParameters,
    JobMlflowRunSchema,
    JobPageSchema,
    JobSchema,
    JobStatusSchema,
)
from .service import (
    RESOURCE_TYPE,
    SEARCHABLE_FIELDS,
    JobIdMlflowrunService,
    JobIdService,
    JobIdStatusService,
    JobService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Job", description="Job endpoint")


@api.route("/")
class JobEndpoint(Resource):
    @inject
    def __init__(self, job_service: JobService, *args, **kwargs) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_service: A JobService object.
        """
        self._job_service = job_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=JobGetQueryParameters, api=api)
    @responds(schema=JobPageSchema, api=api)
    def get(self):
        """Gets a list of all Job resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Job", request_type="GET"
        )
        log.debug("Request received")
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        jobs, total_num_jobs = self._job_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            "jobs",
            build_fn=utils.build_job,
            data=jobs,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_jobs,
            sort_by=sort_by_string,
            descending=descending,
        )


@api.route("/<int:id>")
@api.param("id", "ID for the Job resource.")
class JobIdEndpoint(Resource):
    @inject
    def __init__(self, job_id_service: JobIdService, *args, **kwargs) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdService object.
        """
        self._job_id_service = job_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobSchema, api=api)
    def get(self, id: int):
        """Gets a Job resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Job", request_type="GET", id=id
        )
        job = cast(
            models.Job,
            self._job_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_job(job)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a Job resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Job", request_type="DELETE", id=id
        )
        return self._job_id_service.delete(job_id=id, log=log)


@api.route("/<int:id>/status")
@api.param("id", "ID for the Job resource.")
class JobIdStatusEndpoint(Resource):
    @inject
    def __init__(
        self, job_id_status_service: JobIdStatusService, *args, **kwargs
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdStatusService object.
        """
        self._job_id_status_service = job_id_status_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobStatusSchema, api=api)
    def get(self, id: int):
        """Gets a Job resource's status."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Job", request_type="GET", id=id
        )
        return self._job_id_status_service.get(
            job_id=id, error_if_not_found=True, log=log
        )


@api.route("/<int:id>/mlflowRun")
@api.param("id", "ID for the Job resource.")
class JobIdMlflowrunEndpoint(Resource):
    @inject
    def __init__(
        self,
        job_id_mlflowrun_service: JobIdMlflowrunService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the jobs resource.

        All arguments are provided via dependency injection.

        Args:
            job_id_service: A JobIdStatusService object.
        """
        self._job_id_mlflowrun_service = job_id_mlflowrun_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=JobMlflowRunSchema, api=api)
    def get(self, id: int):
        """Gets a Job resource's mlflow run id."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="JobIdMlflowrunEndpoint",
            request_type="GET",
            job_id=id,
        )
        return self._job_id_mlflowrun_service.get(
            job_id=id, error_if_not_found=True, log=log
        )

    @login_required
    @accepts(schema=JobMlflowRunSchema, api=api)
    @responds(schema=JobMlflowRunSchema, api=api)
    def post(self, id: int):
        """Sets the mlflow run id for a Job"""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="JobIdMlflowrunEndpoint",
            request_type="POST",
            job_id=id,
        )
        parsed_obj = request.parsed_obj  # type: ignore
        return self._job_id_mlflowrun_service.create(
            job_id=id,
            mlflow_run_id=parsed_obj["mlflow_run_id"],
            error_if_not_found=True,
            log=log,
        )


JobSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.Job,
    resource_name=RESOURCE_TYPE,
    route_prefix=V1_JOBS_ROUTE,
    searchable_fields=SEARCHABLE_FIELDS,
    page_schema=JobPageSchema,
    build_fn=utils.build_job,
)
JobSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.Job,
    resource_name=RESOURCE_TYPE,
    response_schema=JobSchema,
    build_fn=utils.build_job,
)

JobTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)
JobTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from .coordinates import (
    BoundingBoxCoordinates,
    BoundingBoxesBatchedGrid,
    NumpyBoundingBoxCoordinates,
    NumpyBoundingBoxesBatchedGrid,
    TensorflowBoundingBoxCoordinates,
    TensorflowBoundingBoxesBatchedGrid,
)
from .iou import (
    BoundingBoxesBatchedGridIOU,
    BoundingBoxesIOU,
    TensorflowBoundingBoxesBatchedGridIOU,
    TensorflowBoundingBoxesIOU,
)
from .postprocessing import (
    BoundingBoxesYOLOV1PostProcessing,
    TensorflowBoundingBoxesYOLOV1Confluence,
    TensorflowBoundingBoxesYOLOV1NMS,
)

__all__ = [
    "BoundingBoxCoordinates",
    "BoundingBoxesBatchedGrid",
    "BoundingBoxesBatchedGridIOU",
    "BoundingBoxesIOU",
    "BoundingBoxesYOLOV1PostProcessing",
    "NumpyBoundingBoxCoordinates",
    "NumpyBoundingBoxesBatchedGrid",
    "TensorflowBoundingBoxCoordinates",
    "TensorflowBoundingBoxesBatchedGrid",
    "TensorflowBoundingBoxesBatchedGridIOU",
    "TensorflowBoundingBoxesIOU",
    "TensorflowBoundingBoxesYOLOV1Confluence",
    "TensorflowBoundingBoxesYOLOV1NMS",
]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Protocol

import structlog
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class ComputeWhLossFn(Protocol):
    def __call__(
        self,
        true_bboxes_cell_wh: Tensor,
        pred_bboxes_cell_wh: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        pass


class YOLOV1LocalizationLoss(object):
    def __init__(
        self,
        coord_weight: float = 5.0,
        object_weight: float = 5.0,
        no_object_weight: float = 1.0,
        wh_loss_type: str = "sq_relative_diff",
    ) -> None:
        self._coord_weight = coord_weight
        self._object_weight = object_weight
        self._no_object_weight = no_object_weight
        self._wh_loss_type = wh_loss_type
        self._compute_wh_loss = self._get_wh_loss_fn(wh_loss_type=wh_loss_type)

    def __call__(self, *args, **kwargs) -> Tensor:
        return self.call(*args, **kwargs)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def call(
        self,
        responsible_pred_bboxes_cell_xywh: Tensor,
        pred_object_conf: Tensor,
        responsible_pred_object_conf: Tensor,
        not_responsible_pred_object_conf: Tensor,
        true_bboxes_cell_xywh: Tensor,
        true_object_conf: Tensor,
        true_object: Tensor,
        true_no_object: Tensor,
    ) -> Tensor:
        # Compute xy and wh losses
        xy_loss = self._compute_xy_loss(
            true_bboxes_cell_xy=true_bboxes_cell_xywh[..., :2],
            pred_bboxes_cell_xy=responsible_pred_bboxes_cell_xywh[..., :2],
            true_object=true_object,
        )
        wh_loss = self._compute_wh_loss(
            true_bboxes_cell_wh=true_bboxes_cell_xywh[..., 2:],
            pred_bboxes_cell_wh=responsible_pred_bboxes_cell_xywh[..., 2:],
            true_object=true_object,
        )

        # Compute obj. loss
        obj_loss = self._compute_obj_loss(
            true_object_conf=true_object_conf,
            pred_object_conf=responsible_pred_object_conf,
            true_object=true_object,
        )

        # Compute no obj. loss
        no_obj_loss = self._compute_no_obj_loss(
            pred_object_conf=pred_object_conf,
            not_responsible_pred_object_conf=not_responsible_pred_object_conf,
            true_object=true_object,
            true_no_object=true_no_object,
        )

        return (
            self._coord_weight * (xy_loss + wh_loss)
            + self._object_weight * obj_loss
            + self._no_object_weight * no_obj_loss
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_xy_loss(
        self,
        true_bboxes_cell_xy: Tensor,
        pred_bboxes_cell_xy: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        sse_xy = tf.reduce_sum(
            tf.square(true_bboxes_cell_xy - pred_bboxes_cell_xy), axis=-1
        )
        reduction_axes = tf.range(tf.rank(sse_xy), dtype=tf.int32)[1:]
        sse_xy = tf.reduce_sum(
            tf.expand_dims(true_object, axis=-1) * sse_xy, axis=reduction_axes
        )

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return sse_xy / (num_obj + 1e-6) / 2

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_wh_square_diff_sqrt_loss(
        self,
        true_bboxes_cell_wh: Tensor,
        pred_bboxes_cell_wh: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        sse_wh = tf.reduce_sum(
            tf.square(tf.sqrt(true_bboxes_cell_wh) - tf.sqrt(pred_bboxes_cell_wh)),
            axis=-1,
        )
        reduction_axes = tf.range(tf.rank(sse_wh), dtype=tf.int32)[1:]
        sse_wh = tf.reduce_sum(
            tf.expand_dims(true_object, axis=-1) * sse_wh, axis=reduction_axes
        )

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return sse_wh / (num_obj + 1e-6) / 2

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_wh_square_relative_diff_loss(
        self,
        true_bboxes_cell_wh: Tensor,
        pred_bboxes_cell_wh: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        sse_wh = tf.reduce_sum(
            tf.square(
                (true_bboxes_cell_wh - pred_bboxes_cell_wh)
                / (true_bboxes_cell_wh + 1e-6)
            ),
            axis=-1,
        )
        reduction_axes = tf.range(tf.rank(sse_wh), dtype=tf.int32)[1:]
        sse_wh = tf.reduce_sum(
            tf.expand_dims(true_object, axis=-1) * sse_wh, axis=reduction_axes
        )

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_obj_reduction_axes = tf.range(tf.rank(true_object), dtype=tf.int32)[1:]
        num_obj = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32), axis=num_obj_reduction_axes
        )

        return sse_wh / (num_obj + 1e-6) / 2

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_obj_loss(
        self,
        true_object_conf: Tensor,
        pred_object_conf: Tensor,
        true_object: Tensor,
    ) -> Tensor:
        sq_diff_obj = tf.square(true_object_conf - pred_object_conf)
        reduction_axes = tf.range(tf.rank(sq_diff_obj), dtype=tf.int32)[1:]
        sum_sq_diff_obj_cells = tf.reduce_sum(
            true_object * sq_diff_obj, axis=reduction_axes
        )

        # Count the number of bounding boxes in cells with an object that are
        # responsible for prediction
        num_boxes_from_obj_cells_reduction_axes = tf.range(
            tf.rank(true_object), dtype=tf.int32
        )[1:]
        num_boxes_from_obj_cells = tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32),
            axis=num_boxes_from_obj_cells_reduction_axes,
        )

        return sum_sq_diff_obj_cells / (num_boxes_from_obj_cells + 1e-6) / 2

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _compute_no_obj_loss(
        self,
        pred_object_conf: Tensor,
        not_responsible_pred_object_conf: Tensor,
        true_object: Tensor,
        true_no_object: Tensor,
    ) -> Tensor:
        n_bounding_boxes = tf.cast(tf.shape(pred_object_conf)[-1], tf.float32)

        # Confidence sum of squares for bounding boxes in cells without an object
        sq_diff_boxes_from_no_obj_cells = tf.square(pred_object_conf)
        no_obj_cells_reduction_axes = tf.range(
            tf.rank(sq_diff_boxes_from_no_obj_cells), dtype=tf.int32
        )[1:]
        sum_sq_diff_boxes_from_no_obj_cells = tf.reduce_sum(
            tf.expand_dims(true_no_object, axis=-1) * sq_diff_boxes_from_no_obj_cells,
            axis=no_obj_cells_reduction_axes,
        )

        # Confidence sum of squares for bounding boxes in cells with an object that
        # are not responsible for prediction
        sq_diff_extra_boxes_from_obj_cells = tf.square(not_responsible_pred_object_conf)
        obj_cells_reduction_axes = tf.range(
            tf.rank(sq_diff_extra_boxes_from_obj_cells), dtype=tf.int32
        )[1:]
        sum_sq_diff_extra_boxes_from_obj_cells = tf.reduce_sum(
            tf.expand_dims(true_object, axis=-1) * sq_diff_extra_boxes_from_obj_cells,
            axis=obj_cells_reduction_axes,
        )

        # Count the number of bounding boxes in cells without an object
        num_boxes_from_obj_cells_reduction_axes = tf.range(
            tf.rank(true_object),
            dtype=tf.int32,
        )[1:]
        num_boxes_from_obj_cells = (n_bounding_boxes - 1) * tf.reduce_sum(
            tf.cast(true_object > 0, tf.float32),
            axis=num_boxes_from_obj_cells_reduction_axes,
        )

        # Count the number of bounding boxes in cells with an object that are not
        # responsible for prediction
        num_boxes_from_no_obj_cells_reduction_axes = tf.range(
            tf.rank(true_no_object),
            dtype=tf.int32,
        )[1:]
        num_boxes_from_no_obj_cells = n_bounding_boxes * tf.reduce_sum(
            tf.cast(true_no_object > 0, tf.float32),
            axis=num_boxes_from_no_obj_cells_reduction_axes,
        )

        return (
            (
                sum_sq_diff_boxes_from_no_obj_cells
                + sum_sq_diff_extra_boxes_from_obj_cells
            )
            / (num_boxes_from_obj_cells + num_boxes_from_no_obj_cells + 1e-6)
            / 2
        )

    def _get_wh_loss_fn(self, wh_loss_type: str) -> ComputeWhLossFn:
        wh_loss_fn_registry: dict[str, ComputeWhLossFn] = {
            "sq_relative_diff": self._compute_wh_square_relative_diff_loss,
            "sq_diff_sqrt": self._compute_wh_square_diff_sqrt_loss,
        }
        wh_loss_fn: ComputeWhLossFn | None = wh_loss_fn_registry.get(wh_loss_type)

        if wh_loss_fn is None:
            raise KeyError(
                f"The wh loss type '{wh_loss_type}' is not supported. The following "
                f"types are currently available: {list(wh_loss_fn_registry.keys())}"
            )

        return wh_loss_fn



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path
from typing import Any, Final, cast

import structlog
import yaml
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.task_engine.type_registry import BUILTIN_TYPES

from .type_coercions import (
    BOOLEAN_PARAM_TYPE,
    FLOAT_PARAM_TYPE,
    INTEGER_PARAM_TYPE,
    STRING_PARAM_TYPE,
    coerce_to_type,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

EXPLICIT_GLOBAL_TYPES: Final[set[str]] = {
    STRING_PARAM_TYPE,
    BOOLEAN_PARAM_TYPE,
    INTEGER_PARAM_TYPE,
    FLOAT_PARAM_TYPE,
}
YAML_FILE_ENCODING: Final[str] = "utf-8"
YAML_EXPORT_SETTINGS: Final[dict[str, Any]] = {
    "indent": 2,
    "sort_keys": False,
    "encoding": YAML_FILE_ENCODING,
}


def export_task_engine_yaml(
    entrypoint: models.EntryPoint,
    entry_point_plugin_files: list[models.EntryPointPluginFile],
    plugin_parameter_types: list[models.PluginTaskParameterType],
    base_dir: Path,
    logger: BoundLogger | None = None,
) -> Path:
    """Export an entrypoint's task engine YAML file to a specified directory.

    Args:
        entrypoint: The entrypoint to export.
        entry_point_plugin_files: The entrypoint's plugin files.
        plugin_parameter_types: The latest snapshots of the plugin parameter types
            accessible to the entrypoint.
        base_dir: The directory to export the task engine YAML file to.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The path to the exported task engine YAML file.
    """
    log = logger or LOGGER.new()  # noqa: F841
    task_yaml_path = Path(base_dir, entrypoint.name).with_suffix(".yml")
    task_engine_dict = build_task_engine_dict(
        entrypoint=entrypoint,
        entry_point_plugin_files=entry_point_plugin_files,
        plugin_parameter_types=plugin_parameter_types,
    )

    with task_yaml_path.open("wt", encoding=YAML_FILE_ENCODING) as f:
        yaml.safe_dump(task_engine_dict, f, **YAML_EXPORT_SETTINGS)

    return task_yaml_path


def build_task_engine_dict(
    entrypoint: models.EntryPoint,
    entry_point_plugin_files: list[models.EntryPointPluginFile],
    plugin_parameter_types: list[models.PluginTaskParameterType],
    logger: BoundLogger | None = None,
) -> dict[str, Any]:
    """Build a dictionary representation of a task engine YAML file.

    Args:
        entrypoint: The entrypoint to export.
        entry_point_plugin_files: The entrypoint's plugin files.
        plugin_parameter_types: The latest snapshots of the plugin parameter types
            accessible to the entrypoint.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        A dictionary representation of a task engine YAML file.
    """
    log = logger or LOGGER.new()  # noqa: F841
    tasks, parameter_types = extract_tasks(
        entry_point_plugin_files, plugin_parameter_types=plugin_parameter_types
    )
    parameters = extract_parameters(entrypoint)
    graph = extract_graph(entrypoint)
    return {
        "types": parameter_types,
        "parameters": parameters,
        "tasks": tasks,
        "graph": graph,
    }


def extract_parameters(
    entrypoint: models.EntryPoint,
    logger: BoundLogger | None = None,
) -> dict[str, Any]:
    """Extract the parameters from an entrypoint.

    Args:
        entrypoint: The entrypoint to extract parameters from.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        A dictionary of the entrypoint's parameters.
    """
    log = logger or LOGGER.new()  # noqa: F841
    parameters: dict[str, Any] = {}
    for param in entrypoint.parameters:
        default_value = param.default_value
        parameters[param.name] = {
            "default": coerce_to_type(x=default_value, type_name=param.parameter_type)
        }

        if param.parameter_type in EXPLICIT_GLOBAL_TYPES:
            parameters[param.name]["type"] = (
                _convert_parameter_type_to_task_engine_type(param.parameter_type)
            )

    return parameters


def extract_tasks(
    entry_point_plugin_files: list[models.EntryPointPluginFile],
    plugin_parameter_types: list[models.PluginTaskParameterType],
    logger: BoundLogger | None = None,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Extract the plugin tasks and parameter types from the entrypoint plugin files.

    Args:
        entry_point_plugin_files: The entrypoint's plugin files.
        plugin_parameter_types: The latest snapshots of the plugin parameter types
            accessible to the entrypoint.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        A tuple containing the plugin tasks dictionary and the parameter types
        dictionary.
    """
    log = logger or LOGGER.new()  # noqa: F841

    tasks: dict[str, Any] = {}
    parameter_types: dict[str, Any] = {}
    for entry_point_plugin_file in entry_point_plugin_files:
        plugin = entry_point_plugin_file.plugin
        plugin_file = entry_point_plugin_file.plugin_file

        for task in plugin_file.tasks:
            input_parameters = sorted(
                task.input_parameters, key=lambda x: x.parameter_number
            )
            output_parameters = sorted(
                task.output_parameters, key=lambda x: x.parameter_number
            )

            tasks[task.plugin_task_name] = {
                "plugin": _build_plugin_field(plugin, plugin_file, task),
            }
            if input_parameters:
                tasks[task.plugin_task_name]["inputs"] = _build_task_inputs(
                    input_parameters
                )

            if output_parameters:
                tasks[task.plugin_task_name]["outputs"] = _build_task_outputs(
                    output_parameters
                )

            for param in input_parameters + output_parameters:
                name = param.parameter_type.name
                if name not in BUILTIN_TYPES:
                    parameter_types[name] = param.parameter_type.structure

            # HACK: THIS IS A WORKAROUND THAT VIOLATES IDEMPOTENCE/REPRODUCIBILITY!
            #
            # This workaround allows users to create and use parameter types that are
            # only used indirectly, such as when defining a structured parameter type.
            # This is a "hack" because the objects in `plugin_parameter_types` are the
            # latest available snapshots, not the snapshots that were associated with
            # the plugins when the entrypoint was saved/updated. This is in contrast
            # with the parameter types accumulated in the previous for loop block, which
            # are linked to the entrypoint and job by their snapshot id instead of the
            # resource id. This difference means that the task engine YAML files are not
            # 100% reproducible, as any changes to an "indirect" plugin parameter type
            # will be immediately reflected in subsequent download requests made to the
            # job files workflow.
            for parameter_type in plugin_parameter_types:
                name = parameter_type.name
                if name not in BUILTIN_TYPES and name not in parameter_types:
                    parameter_types[name] = parameter_type.structure

    return tasks, parameter_types


def extract_graph(
    entrypoint: models.EntryPoint,
    logger: BoundLogger | None = None,
) -> dict[str, Any]:
    """Extract the task graph from an entrypoint.

    Args:
        entrypoint: The entrypoint containing the task graph.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        A dictionary representation of the entrypoint's task graph.
    """
    log = logger or LOGGER.new()  # noqa: F841
    return cast(dict[str, Any], yaml.safe_load(entrypoint.task_graph))


def _build_plugin_field(
    plugin: models.Plugin, plugin_file: models.PluginFile, task: models.PluginTask
) -> str:
    if plugin_file.filename == "__init__.py":
        # Omit filename from plugin import path if it is an __init__.py file.
        module_parts = [Path(x).stem for x in Path(plugin_file.filename).parts[:-1]]

    else:
        module_parts = [Path(x).stem for x in Path(plugin_file.filename).parts]

    return ".".join([plugin.name, *module_parts, task.plugin_task_name])


def _build_task_inputs(
    input_parameters: list[models.PluginTaskInputParameter],
) -> list[dict[str, Any]]:
    return [
        {
            "name": input_param.name,
            "type": input_param.parameter_type.name,
            "required": input_param.required,
        }
        for input_param in input_parameters
    ]


def _build_task_outputs(
    output_parameters: list[models.PluginTaskOutputParameter],
) -> list[dict[str, Any]] | dict[str, Any]:
    if len(output_parameters) == 1:
        return {output_parameters[0].name: output_parameters[0].parameter_type.name}

    return [
        {output_param.name: output_param.parameter_type.name}
        for output_param in output_parameters
    ]


def _convert_parameter_type_to_task_engine_type(parameter_type: str) -> Any:
    conversion_map = {
        "boolean": "boolean",
        "string": "string",
        "float": "number",
        "integer": "integer",
        "list": {"list": "any"},
        "mapping": {"mapping": ["string", "any"]},
    }
    return conversion_map[parameter_type]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Test suite for plugin parameter type operations.

This module contains a set of tests that validate the CRUD operations and additional
functionalities for the plugin parameter type entity. The tests ensure that the plugin
parameter types can be submitted and retrieved as expected through the REST API.
"""
from typing import Any

import pytest
from flask.testing import FlaskClient
from flask_sqlalchemy import SQLAlchemy
from werkzeug.test import TestResponse

from dioptra.restapi.routes import V1_PLUGIN_PARAMETER_TYPES_ROUTE, V1_ROOT

from ..lib import actions, asserts, helpers

# -- Actions ---------------------------------------------------------------------------


def modify_plugin_parameter_type(
    client: FlaskClient,
    id: int,
    new_name: str,
    new_structure: dict[str, Any] | None = None,
    new_description: str | None = None,
) -> TestResponse:
    """Rename a Plugin Parameter Type using the API.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to rename.
        new_name: The new name to assign to the plugin parameter type.
        new_structure: The new structure to assign to the plugin parameter type.
        new_description: The new description to assign to the plugin parameter type.

    Returns:
        The response from the API.
    """
    json_payload = {"name": new_name}

    if new_structure is not None:
        json_payload["structure"] = new_structure

    if new_description is not None:
        json_payload["description"] = new_description

    return client.put(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        json=json_payload,
        follow_redirects=True,
    )


def delete_plugin_parameter_type_with_id(client: FlaskClient, id: int) -> TestResponse:
    """Delete a plugin parameter type using the API.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to delete.

    Returns:
        The response from the API.
    """
    return client.delete(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        follow_redirects=True,
    )


# -- Assertions ------------------------------------------------------------------------


def assert_retrieving_plugin_parameter_types_works(
    client: FlaskClient,
    expected: list[dict[str, Any]],
    group_id: int | None = None,
    search: str | None = None,
    paging_info: dict[str, Any] | None = None,
) -> None:
    """Assert that retrieving plugin parameter types works.

    Args:
        client: The Flask test client.
        expected: The expected response from the API.
        group_id: The group of the plugin parameter type.
        search: The search query parameters.
        paging_info: The paging query parameters.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    if group_id is not None:
        query_string["groupId"] = group_id

    if search is not None:
        query_string["search"] = search

    if paging_info is not None:
        query_string["index"] = paging_info["index"]
        query_string["pageLength"] = paging_info["page_length"]

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["data"] == expected


def assert_sorting_plugin_parameter_type_works(
    client: FlaskClient,
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Assert that plugin parameter types can be sorted by column ascending/descending.

    Args:
        client: The Flask test client.
        expected: The expected order of type ids after sorting.
          See test_plugin_parameter_type_sort for expected orders.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """

    query_string: dict[str, Any] = {}

    query_string["sortBy"] = sortBy
    query_string["descending"] = descending

    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}",
        query_string=query_string,
        follow_redirects=True,
    )

    response_data = response.get_json()
    # remove plugin param types created by default before testing
    names_to_remove = ["any", "string", "integer", "number", "boolean", "null"]
    filtered_data = [
        item for item in response_data["data"] if item["name"] not in names_to_remove
    ]

    param_ids = [param["id"] for param in filtered_data]

    assert response.status_code == 200 and param_ids == expected


def assert_retrieving_plugin_parameter_type_by_id_works(
    client: FlaskClient, id: int, expected: dict[str, Any]
) -> None:
    """Assert that retrieving a plugin parameter type by id works.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_plugin_parameter_type_name_matches_expected_name(
    client: FlaskClient, id: int, expected_name: str
) -> None:
    """Assert that the name of a plugin parameter type matches the expected name.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to retrieve.
        expected_name: The expected name of the plugin parameter type.

    Raises:
        AssertionError: If the response status code is not 200 or if the name of the
            experiment does not match the expected name.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        follow_redirects=True,
    )
    assert response.status_code == 200 and response.get_json()["name"] == expected_name


def assert_updating_plugin_parameter_type_works(
    client: FlaskClient, id: int, expected: dict[str, Any]
) -> None:
    """Assert that updating a plugin parameter type by id works.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to retrieve.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 200 or if the API response
            does not match the expected response.
    """
    response = client.put(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_deleting_plugin_parameter_type_by_id_works(
    client: FlaskClient, id: int, expected: dict[str, Any]
) -> None:
    """Assert that deleting a plugin parameter type by id works.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to delete.
        expected: The expected response from the API.

    Raises:
        AssertionError: If the response status code is not 404 or if the API response
            does not match the expected response.
    """
    response = client.delete(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}", follow_redirects=True
    )
    assert response.status_code == 200 and response.get_json() == expected


def assert_plugin_parameter_type_is_not_found(client: FlaskClient, id: int) -> None:
    """Assert that a plugin parameter type is not found.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to retrieve.

    Raises:
        AssertionError: If the response status code is not 404.
    """
    response = client.get(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        follow_redirects=True,
    )
    assert response.status_code == 404


def assert_cannot_rename_invalid_plugin_parameter_type(
    client: FlaskClient, id: int, json_payload: dict[str, Any]
) -> None:
    """Assert that attempting to rename a Plugin Parameter Type with invalid
    parameters using the API fails.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to rename.
        json_payload: The full payload to be sent to the API.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = client.put(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        json=json_payload,
        follow_redirects=True,
    )
    assert response.status_code == 400


def assert_cannot_create_invalid_plugin_parameter_type(
    client: FlaskClient, json_payload: dict[str, Any]
) -> None:
    """Assert that attempting to create a Plugin Parameter Type with invalid
    parameters using the API fails.

    Args:
        client: The Flask test client.
        json_payload: The full payload to be sent to the API.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = client.post(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}",
        json=json_payload,
        follow_redirects=True,
    )
    assert response.status_code == 400


def assert_cannot_create_existing_plugin_parameter_type(
    client: FlaskClient, name: str, group_id: int
) -> None:
    """Assert that attempting to create a Plugin Parameter Type with an existing
    name using the API fails.

    Args:
        client: The Flask test client.
        name: The name of the plugin parameter type.
        group_id: The group of the plugin parameter type.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = actions.register_plugin_parameter_type(
        client, name=name, group_id=group_id, structure=dict(), description=""
    )
    assert response.status_code == 400


def assert_cannot_rename_plugin_parameter_type_to_existing_name(
    client: FlaskClient,
    id: int,
    existing_name: str,
    new_structure: dict[str, Any],
    new_description: str,
) -> None:
    """Assert that attempting to rename a Plugin Parameter Type to an existing
    name using the API fails.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to rename.
        existing_name: The name of an existing plugin parameter type.
        new_structure: The new structure to assign to the plugin parameter type.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = modify_plugin_parameter_type(
        client,
        id=id,
        new_name=existing_name,
        new_structure=new_structure,
        new_description=new_description,
    )
    assert response.status_code == 409


def assert_cannot_delete_invalid_plugin_parameter_type(
    client: FlaskClient, id: Any, json_payload: dict[str, Any]
) -> None:
    """Assert that attempting to delete a Plugin Parameter Type with invalid
    parameters using the API fails.

    Args:
        client: The Flask test client.
        id: The id of the plugin parameter type to delete.
        json_payload: The full payload to be sent to the API.

    Raises:
        AssertionError: If the response status code is not 400.
    """
    response = client.delete(
        f"/{V1_ROOT}/{V1_PLUGIN_PARAMETER_TYPES_ROUTE}/{id}",
        json=json_payload,
        follow_redirects=True,
    )
    assert response.status_code == 400


def assert_plugin_parameter_type_content_matches_expectations(
    response: dict[str, Any], expected_contents: dict[str, Any]
) -> None:
    """Assert that a Plugin Parameter Type's fields match the expected values.

    Args:
        response: The response to check.
        expected_contents: The expected values for the response.
    """
    expected_keys = {
        "id",
        "snapshot",
        "group",
        "user",
        "createdOn",
        "lastModifiedOn",
        "latestSnapshot",
        "hasDraft",
        "name",
        "structure",
        "description",
        "tags",
    }
    assert set(response.keys()) == expected_keys

    # Validate the non-Ref fields
    assert isinstance(response["id"], int)
    assert isinstance(response["snapshot"], int)
    assert isinstance(response["name"], str)
    assert isinstance(response["structure"], (dict, type(None)))
    assert isinstance(response["description"], (str, type(None)))
    assert isinstance(response["createdOn"], str)
    assert isinstance(response["lastModifiedOn"], str)
    assert isinstance(response["latestSnapshot"], bool)
    assert isinstance(response["hasDraft"], bool)

    assert response["name"] == expected_contents["name"]
    assert response["structure"] == expected_contents["structure"]
    assert response["description"] == expected_contents["description"]

    assert helpers.is_iso_format(response["createdOn"])
    assert helpers.is_iso_format(response["lastModifiedOn"])

    # Validate the UserRef structure
    assert isinstance(response["user"]["id"], int)
    assert isinstance(response["user"]["username"], str)
    assert isinstance(response["user"]["url"], str)
    assert response["user"]["id"] == expected_contents["user_id"]

    # Validate the GroupRef structure
    assert isinstance(response["group"]["id"], int)
    assert isinstance(response["group"]["name"], str)
    assert isinstance(response["group"]["url"], str)
    assert response["group"]["id"] == expected_contents["group_id"]


# -- Tests -----------------------------------------------------------------------------


def test_get_all_plugin_parameter_types(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that all Plugin Parameter Types can be retrieved.

        Scenario: Get a List of Plugin Parameter Types
            Given I am an authorized user,
            I need to be able to submit a GET request with optional query parameters
            in order to retrieve a list of plugin parameter types matching the
            parameters.

    This test validates this scenario by following these actions:

    - The user is able to retrieve a list of all plugin parameter types.
    - The returned list of plugin parameter types match the full list of existing plugin
      parameter types.
    """
    plugin_param_type_expected_list = list(registered_plugin_parameter_types.values())
    assert_retrieving_plugin_parameter_types_works(
        client, expected=plugin_param_type_expected_list
    )


@pytest.mark.parametrize(
    "sortBy, descending , expected",
    [
        (
            None,
            None,
            ["plugin_param_type1", "plugin_param_type2", "plugin_param_type3"],
        ),
        (
            "name",
            True,
            ["plugin_param_type2", "plugin_param_type3", "plugin_param_type1"],
        ),
        (
            "name",
            False,
            ["plugin_param_type1", "plugin_param_type3", "plugin_param_type2"],
        ),
        (
            "createdOn",
            True,
            ["plugin_param_type3", "plugin_param_type2", "plugin_param_type1"],
        ),
        (
            "createdOn",
            False,
            ["plugin_param_type1", "plugin_param_type2", "plugin_param_type3"],
        ),
    ],
)
def test_plugin_parameter_type_sort(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
    sortBy: str,
    descending: bool,
    expected: list[str],
) -> None:
    """Test that plugin param types can be sorted by column.

    Given an authenticated user and registered types, this test validates the following
    sequence of actions:

    - A user registers three types, "image_shape", "model_output", "model".
    - The user is able to retrieve a list of all registered types sorted by a column
      ascending/descending.
    - The returned list of plugin param types matches the order in the parametrize lists
      above.
    """

    expected_ids = [
        registered_plugin_parameter_types[expected_name]["id"]
        for expected_name in expected
    ]
    assert_sorting_plugin_parameter_type_works(
        client, sortBy, descending, expected=expected_ids
    )


def test_plugin_parameter_type_search_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that multiple Plugin Parameter Types can be retrieved using search terms.

    Given an authenticated user and registered plugin parameter types, this test
    validates this scenario by following these actions:

    - The user is able to retrieve 2 lists of the submitted plugin parameter types
      using different query parameters.
    - The returned lists of plugin parameter types match the searches provided
      during both submissions.
    """
    assert_retrieving_plugin_parameter_types_works(
        client,
        expected=[registered_plugin_parameter_types["string"]],
        search="name:string",
    )
    assert_retrieving_plugin_parameter_types_works(
        client,
        expected=[
            registered_plugin_parameter_types["plugin_param_type2"],
            registered_plugin_parameter_types["plugin_param_type3"],
        ],
        search="*model*",
    )
    plugin_param_type_expected_list = list(registered_plugin_parameter_types.values())
    assert_retrieving_plugin_parameter_types_works(
        client, expected=plugin_param_type_expected_list, search="*"
    )


def test_plugin_parameter_type_group_query(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that Plugin Parameter Types can be retrieved using a group filter.

    Given an authenticated user and registered plugin parameter types, this test
    validates this scenario by following these actions:

    - The user is able to retrieve a list of all submitted plugin parameter types
      owned by the default group.
    - The returned list of plugin parameter types matches the expected list owned by
      the default group.
    """
    plugin_param_type_expected_list = list(registered_plugin_parameter_types.values())
    assert_retrieving_plugin_parameter_types_works(
        client,
        expected=plugin_param_type_expected_list,
        group_id=auth_account["groups"][0]["id"],
    )


def test_create_plugin_parameter_type(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> None:
    """Test that Plugin Parameter Types can be created using the API.

        Scenario: Create a Plugin Parameter Type
            Given I am an authorized user,
            I need to be able to submit a POST request with body parameters name
            group_id, and optional structure
            in order to create a new plugin parameter type.

    This test validates the following sequence of actions:
    - A user submits a POST request for a plugin parameter type.
    - The user retrieves information about one plugin parameter type using its id.
    - The returned information matches what was provided during creation.
    - This set of steps is repeated when testing using a missing parameter, extra
      parameter and a parameter with the wrong type during the POST request.
    - The second, third and fourth POST operations are expected to fail.
    """
    name = "test_plugin_param_type"
    user_id = auth_account["id"]
    group_id = auth_account["groups"][0]["id"]
    description = "This is a plugin parameter type"
    plugin_param_type1_response = actions.register_plugin_parameter_type(
        client,
        name=name,
        group_id=group_id,
        description=description,
    )
    plugin_param_type1_expected = plugin_param_type1_response.get_json()
    assert_plugin_parameter_type_content_matches_expectations(
        response=plugin_param_type1_expected,
        expected_contents={
            "name": name,
            "user_id": user_id,
            "group_id": group_id,
            "description": description,
            "structure": None,
        },
    )
    assert_retrieving_plugin_parameter_type_by_id_works(
        client,
        id=plugin_param_type1_expected["id"],
        expected=plugin_param_type1_expected,
    )

    # Error case (missing parameter)
    assert_cannot_create_invalid_plugin_parameter_type(client, json_payload={})

    # Error case (extra parameter)
    assert_cannot_create_invalid_plugin_parameter_type(
        client,
        json_payload={
            "name": "plugin_param_type_2",
            "extra_param": "invalid",
            "extra_param2": "invalid2",
        },
    )

    # Error case (wrong type)
    assert_cannot_create_invalid_plugin_parameter_type(
        client,
        json_payload={
            "name": "plugin_param_type_1",
            "group_id": "invalid",
        },
    )


def test_get_plugin_parameter_type(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that a Plugin Parameter Type can be retrieved following the scenario below:

        Scenario: Get a Plugin Parameter Type using its ID
            Given I am an authorized user and at least one plugin parameter type exists,
            I need to be able to submit a GET request with path parameter id
            in order to retrieve the plugin parameter type with the specified ID.

    This test validates this scenario by following these actions:

    - A user submits a POST request for two plugin parameter types.
    - The user retrieves information about one plugin parameter type using its id.
    - The returned information matches what was provided during creation.
    """
    plugin_param_type1_expected = registered_plugin_parameter_types[
        "plugin_param_type1"
    ]

    # Get a Plugin Parameter Type using its ID
    assert_retrieving_plugin_parameter_type_by_id_works(
        client,
        id=plugin_param_type1_expected["id"],
        expected=plugin_param_type1_expected,
    )


def test_delete_plugin_parameter_type(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that a Plugin Parameter Type can be deleted following the scenario below:

         Scenario: Delete a Plugin Parameter Type
              Given I am an authorized user and at least one plugin parameter type
              exists, I need to be able to submit a DELETE request with path parameter
              id in order to delete a plugin parameter type with the specified ID.

    This test validates these scenarios by following these actions:

     - A user submits a POST request for two plugin parameter types.
     - The user retrieves information about one plugin parameter type using its id.
     - The returned information matches what was provided during creation.
     - The plugin parameter type is then deleted.
     - A search by id for the deleted plugin parameter type should return a 404 error.
    """
    plugin_param_type1 = registered_plugin_parameter_types["plugin_param_type1"]

    # Delete a Plugin Parameter Type using its ID
    assert_retrieving_plugin_parameter_type_by_id_works(
        client, id=plugin_param_type1["id"], expected=plugin_param_type1
    )
    delete_plugin_parameter_type_with_id(client, plugin_param_type1["id"])
    assert_plugin_parameter_type_is_not_found(client, plugin_param_type1["id"])


def test_modify_plugin_parameter_type(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that a Plugin Parameter Type can be modified following the scenario below:

        Scenario: Modify an Existing Plugin Parameter Type
            Given I am an authorized user and at least one plugin parameter type exists,
            I need to be able to submit a PUT request with path parameter id and body
            parameters name and structure in order to update the plugin parameter type
            with the new name and structure parameters.

    This test validates this scenario by following these actions:

    - A user creates a plugin parameter type.
    - The user is able to retrieve information about the plugin parameter type that
      matches the information that was provided during creation.
    - The user modifies this same plugin parameter type.
    - The user retrieves information about the same plugin parameter type and it
      reflects the changes.
    - A user then attempts to modify a plugin parameter type with a missing parameter,
      extra parameter, and a parameter with the wrong type in the request.
    - Each of these cases causes an error to be returned.
    """
    plugin_param_type1 = registered_plugin_parameter_types["plugin_param_type1"]
    plugin_param_type2 = registered_plugin_parameter_types["plugin_param_type2"]
    plugin_param_type3 = registered_plugin_parameter_types["plugin_param_type3"]

    start_name = plugin_param_type1["name"]
    updated_name = "changed_name"
    updated_structure = {"key": "value"}

    # Modify an Existing Plugin Parameter Type
    assert_plugin_parameter_type_name_matches_expected_name(
        client, id=plugin_param_type1["id"], expected_name=start_name
    )
    modify_plugin_parameter_type(
        client,
        id=plugin_param_type1["id"],
        new_name=updated_name,
        new_structure=updated_structure,
        new_description=plugin_param_type1["description"],
    )
    assert_plugin_parameter_type_name_matches_expected_name(
        client, id=plugin_param_type1["id"], expected_name=updated_name
    )

    # Error case (missing parameter)
    assert_cannot_rename_invalid_plugin_parameter_type(
        client, id=plugin_param_type2["id"], json_payload={}
    )

    # Error case (extra parameter)
    assert_cannot_rename_invalid_plugin_parameter_type(
        client,
        id=plugin_param_type2["id"],
        json_payload={
            "name": "param_name",
            "extra_param": "invalid",
        },
    )

    # Error case (wrong type)
    assert_cannot_rename_invalid_plugin_parameter_type(
        client,
        id=plugin_param_type2["id"],
        json_payload={"name": 42, "structure": dict()},
    )

    # Attempt to rename a Plugin Parameter Type to an existing name
    assert_cannot_rename_plugin_parameter_type_to_existing_name(
        client,
        id=plugin_param_type1["id"],
        existing_name=plugin_param_type3["name"],
        new_structure=None,
        new_description=None,
    )


def test_cannot_create_existing_plugin_parameter_type(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that creating an existing Plugin Parameter Type produces an error.

    This test validates the following sequence of actions:
    - A user attempts to create a plugin parameter type with a name that already
      exists.
    - This causes a 400 error to be returned.
    """
    existing_plugin_param_type = registered_plugin_parameter_types["plugin_param_type1"]

    assert_cannot_create_existing_plugin_parameter_type(
        client,
        name=existing_plugin_param_type,
        group_id=auth_account["groups"][0]["id"],
    )


def test_cannot_retrieve_nonexistent_plugin_parameter_type(
    client: FlaskClient, db: SQLAlchemy, auth_account: dict[str, Any]
) -> None:
    """Test that retrieving an nonexistent Plugin Parameter Type produces an error.

    This test validates the following sequence of actions:
    - A user attempts to retrieve a plugin parameter type that doesn't exist.
    - This causes a 404 error to be returned.
    """
    assert_plugin_parameter_type_is_not_found(client, id=42)


def test_manage_existing_plugin_parameter_type_draft(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that a draft of an existing plugin parameter type can be created
    and managed by the user
    Given an authenticated user and registered plugin parameter types, this
    test validates the following sequence of actions:
    - The user creates a draft of an existing plugin parameter type
    - The user retrieves information about the draft and gets the expected response
    - The user attempts to create another draft of the same existing plugin
      parameter type
    - The request fails with an appropriate error message and response code.
    - The user modifies the name of the plugin parameter type in the draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    plugin_param_type = registered_plugin_parameter_types["plugin_param_type1"]
    name = "draft"
    new_name = "draft2"
    description = "description"

    # test creation
    payload = {"name": name, "description": description, "structure": None}
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin_param_type["group"]["id"],
        "resource_id": plugin_param_type["id"],
        "resource_snapshot_id": plugin_param_type["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.create_existing_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)
    asserts.assert_retrieving_draft_by_resource_id_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        expected=response,
    )
    asserts.assert_creating_another_existing_draft_fails(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )

    # test modification
    payload = {"name": new_name, "description": description, "structure": None}
    expected = {
        "user_id": auth_account["id"],
        "group_id": plugin_param_type["group"]["id"],
        "resource_id": plugin_param_type["id"],
        "resource_snapshot_id": plugin_param_type["snapshot"],
        "num_other_drafts": 0,
        "payload": payload,
    }
    response = actions.modify_existing_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        payload=payload,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(response, expected)

    # test deletion
    actions.delete_existing_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )
    asserts.assert_existing_draft_is_not_found(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )


def test_manage_new_plugin_parameter_type_drafts(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
) -> None:
    """Test that drafts of plugin parameter type can be created and managed by the user
    Given an authenticated user, this test validates the following sequence of actions:
    - The user creates two plugin parameter type drafts
    - The user retrieves information about the drafts and gets the expected response
    - The user modifies the description of the plugin parameter type in the first draft
    - The user retrieves information about the draft and gets the expected response
    - The user deletes the first draft
    - The user attempts to retrieve information about the deleted draft.
    - The request fails with an appropriate error message and response code.
    """
    group_id = auth_account["groups"][0]["id"]
    drafts = {
        "draft1": {
            "name": "plugin_parameter_type1",
            "description": "new plugin parameter type",
            "structure": None,
        },
        "draft2": {
            "name": "plugin_parameter_type2",
            "description": None,
            "structure": None,
        },
    }

    # test creation
    draft1_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft1"],
    }
    draft1_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        group_id=group_id,
        payload=drafts["draft1"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft1_response, draft1_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        draft_id=draft1_response["id"],
        expected=draft1_response,
    )
    draft2_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": drafts["draft2"],
    }
    draft2_response = actions.create_new_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        group_id=group_id,
        payload=drafts["draft2"],
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        draft2_response, draft2_expected
    )
    asserts.assert_retrieving_draft_by_id_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        draft_id=draft2_response["id"],
        expected=draft2_response,
    )
    asserts.assert_retrieving_drafts_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        expected=[draft1_response, draft2_response],
    )

    # test modification
    draft1_mod = {"name": "draft1", "description": "new description", "structure": None}
    draft1_mod_expected = {
        "user_id": auth_account["id"],
        "group_id": group_id,
        "payload": draft1_mod,
    }
    response = actions.modify_new_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        draft_id=draft1_response["id"],
        payload=draft1_mod,
    ).get_json()
    asserts.assert_draft_response_contents_matches_expectations(
        response, draft1_mod_expected
    )

    # test deletion
    actions.delete_new_resource_draft(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        draft_id=draft1_response["id"],
    )
    asserts.assert_new_draft_is_not_found(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        draft_id=draft1_response["id"],
    )


def test_manage_plugin_parameter_type_snapshots(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
) -> None:
    """Test that different snapshots of a plugin parameter type can be retrieved
    by the user.

    Given an authenticated user and registered plugin parameter types, this test
    validates the following sequence of actions:

    - The user modifies a plugin parameter type
    - The user retrieves information about the original snapshot of the plugin
      parameter type and gets the expected response
    - The user retrieves information about the new snapshot of the plugin
      parameter type and gets the expected response
    - The user retrieves a list of all snapshots of the plugin parameter type
      and gets the expected response
    """
    plugin_param_type_to_rename = registered_plugin_parameter_types[
        "plugin_param_type1"
    ]
    modified_plugin_param_type = modify_plugin_parameter_type(
        client,
        id=plugin_param_type_to_rename["id"],
        new_name=plugin_param_type_to_rename["name"] + "modified",
        new_description=plugin_param_type_to_rename["description"],
        new_structure=plugin_param_type_to_rename["structure"],
    ).get_json()
    modified_plugin_param_type.pop("hasDraft")
    plugin_param_type_to_rename.pop("hasDraft")
    plugin_param_type_to_rename["latestSnapshot"] = False
    plugin_param_type_to_rename["lastModifiedOn"] = modified_plugin_param_type[
        "lastModifiedOn"
    ]
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type_to_rename["id"],
        snapshot_id=plugin_param_type_to_rename["snapshot"],
        expected=plugin_param_type_to_rename,
    )
    asserts.assert_retrieving_snapshot_by_id_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=modified_plugin_param_type["id"],
        snapshot_id=modified_plugin_param_type["snapshot"],
        expected=modified_plugin_param_type,
    )
    expected_snapshots = [plugin_param_type_to_rename, modified_plugin_param_type]
    asserts.assert_retrieving_snapshots_works(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type_to_rename["id"],
        expected=expected_snapshots,
    )


def test_tag_plugin_parameter_type(
    client: FlaskClient,
    db: SQLAlchemy,
    auth_account: dict[str, Any],
    registered_plugin_parameter_types: dict[str, Any],
    registered_tags: dict[str, Any],
) -> None:
    """Test that tags can applied to plugin parameter types.

    Given an authenticated user and registered plugin parameter types, this test
    validates the following sequence of actions:
    """
    plugin_param_type = registered_plugin_parameter_types["plugin_param_type1"]
    tags = [tag["id"] for tag in registered_tags.values()]

    # test append
    response = actions.append_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        tag_ids=[tags[0], tags[1]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1]]
    )
    response = actions.append_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[1], tags[2]]
    )

    # test remove
    actions.remove_tag(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        tag_id=tags[1],
    )
    response = actions.get_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[0], tags[2]]
    )

    # test modify
    response = actions.modify_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
        tag_ids=[tags[1], tags[2]],
    )
    asserts.assert_tags_response_contents_matches_expectations(
        response.get_json(), [tags[1], tags[2]]
    )

    # test delete
    response = actions.remove_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )
    response = actions.get_tags(
        client,
        resource_route=V1_PLUGIN_PARAMETER_TYPES_ROUTE,
        resource_id=plugin_param_type["id"],
    )
    asserts.assert_tags_response_contents_matches_expectations(response.get_json(), [])



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Implement static type validation for a declarative experiment description.
"""

import functools
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Optional

from dioptra.task_engine import type_registry, types, util
from dioptra.task_engine.issues import IssueSeverity, IssueType, ValidationIssue

# Type aliases for type annotations
_TypeMap = Mapping[str, types.Type]
# Map from a name to a definition of something.  Applicable to types, tasks,
# parameters, graph steps.
_DefMap = Mapping[str, Any]


def _get_reference_type(
    reference: str,
    type_reg: _TypeMap,
    global_parameter_types: _TypeMap,
    graph: _DefMap,
    tasks: _DefMap,
) -> types.Type:
    """
    Given a reference, look up its type.

    Args:
        reference: A reference, without the leading "$"
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to
            Type instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short name to task definition

    Returns:
        A type
    """

    if "." in reference:
        ref_name, ref_output_name = reference.split(".", 1)

    else:
        ref_name = reference
        ref_output_name = None

    # Lots of correctness assumptions inherent below.  E.g. that other
    # validation has previously occurred, to catch things like bad references.
    if ref_output_name is None and ref_name in global_parameter_types:
        type_ = global_parameter_types[ref_name]

    else:
        step_def = graph[ref_name]
        task_short_name = util.step_get_plugin_short_name(step_def)

        # Assume well-formed step definition (see the comment about
        # correctness assumptions above... but we still have to satisfy mypy)
        assert task_short_name is not None

        task_def = tasks[task_short_name]
        task_outputs = task_def.get("outputs", [])

        if not isinstance(task_outputs, list):
            task_outputs = [task_outputs]

        output_type_name = None
        if ref_output_name is None:
            output_def = task_outputs[0]
            output_type_name = next(iter(output_def.values()))
        else:
            for output_def in task_outputs:
                this_output_name, output_type_name = next(iter(output_def.items()))
                if this_output_name == ref_output_name:
                    break

        # Assume references have already been validated, so we will always
        # find an output which matches the reference.
        assert output_type_name
        type_ = type_reg[output_type_name]

    return type_


def _find_common_base_of_two_types(type1: types.Type, type2: types.Type) -> types.Type:
    """
    Find the most specific common base type between the pair of types given.
    Will return the "any" type if nothing more specific is found.

    Args:
        type1: A Type instance
        type2: A Type instance

    Returns:
        A Type instance
    """
    common_base: Optional[types.Type]

    if isinstance(type1, types.SimpleType) and isinstance(type2, types.SimpleType):
        if type1 == type2:
            common_base = type1
        elif type1.is_subtype_of(type2):
            common_base = type2
        elif type2.is_subtype_of(type1):
            common_base = type1
        else:
            # Try to find a base of type1 which is also a base of type2.
            common_base = type1.super_type
            while common_base is not None:
                if type2.is_subtype_of(common_base):
                    break
                common_base = common_base.super_type
            else:
                # Can't find a common base, so use any.
                common_base = type_registry.TYPE_ANY

    else:
        common_base = type_registry.TYPE_ANY

    return common_base


def _find_common_base_type(types_: Iterable[types.Type]):
    """
    Find the most specific common base type among all of the given types.  Will
    return the "any" type if nothing more specific is found.

    Args:
        types_: An iterable of Type instances

    Returns:
        A Type instance
    """
    try:
        common_base = functools.reduce(_find_common_base_of_two_types, types_)
    except TypeError as e:
        raise ValueError("_find_common_base_type() requires at least one type") from e

    return common_base


def _infer_type_from_mapping(
    mapping: Mapping,
    type_reg: Optional[_TypeMap] = None,
    global_parameter_types: Optional[_TypeMap] = None,
    graph: Optional[_DefMap] = None,
    tasks: Optional[_DefMap] = None,
) -> types.Type:
    """
    Infer the type of the given Mapping value.  This always produces a mapping
    type.  If the latter four parameters are non-None, reference resolution is
    enabled.  That means if the value contains strings which look like a
    reference, they are looked up and resolved to types.  If reference
    resolution is disabled, the string type will be inferred for all string
    values, no matter what they look like.

    Args:
        mapping: The mapping value to infer a type for
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to
            Type instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short name to task definition

    Returns:
        The value type
    """
    type_: types.Type

    if mapping:
        key_types = (
            # disable reference resolution here
            _infer_type(key)
            for key in mapping
        )

        key_base_type = _find_common_base_type(key_types)

        if key_base_type is type_registry.TYPE_STRING:
            # If keys are strings, infer an enumerated mapping type
            prop_types = {
                prop_name: _infer_type(
                    prop_value, type_reg, global_parameter_types, graph, tasks
                )
                for prop_name, prop_value in mapping.items()
            }

            type_ = types.StructuredType(
                types.TypeStructure(types.StructureType.MAPPING, prop_types)
            )

        elif key_base_type is type_registry.TYPE_INTEGER:
            # Integer-keyed keytype/valuetype style mappings are also
            # permitted.  Need to come up with a single type from potentially
            # multiple different property value types.
            value_types = {
                _infer_type(v, type_reg, global_parameter_types, graph, tasks)
                for v in mapping.values()
            }

            if len(value_types) == 1:
                value_type = next(iter(value_types))
            else:
                value_type = types.UnionType(value_types)

            type_ = types.StructuredType(
                types.TypeStructure(
                    types.StructureType.MAPPING, [key_base_type, value_type]
                )
            )

        else:
            # Else, there is no mapping type in the type system which
            # supports the combination of key and value types present in
            # the mapping.  So we must infer any.
            type_ = type_registry.TYPE_ANY

    else:
        # If an empty mapping, mapping[{}] is inferred (i.e. enumerated
        # property mapping with no properties).
        type_ = types.StructuredType(
            types.TypeStructure(types.StructureType.MAPPING, {})
        )

    return type_


def _infer_type_from_iterable(
    iterable: Iterable,
    type_reg: Optional[_TypeMap] = None,
    global_parameter_types: Optional[_TypeMap] = None,
    graph: Optional[_DefMap] = None,
    tasks: Optional[_DefMap] = None,
) -> types.Type:
    """
    Infer the type of the given iterable value.  This always produces a tuple
    type.  If the latter four parameters are non-None, reference resolution is
    enabled.  That means if the value contains strings which look like a
    reference, they are looked up and resolved to types.  If reference
    resolution is disabled, the string type will be inferred for all string
    values, no matter what they look like.

    Args:
        iterable: The iterable value to infer a type for
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to
            Type instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short name to task definition

    Returns:
        The value type
    """
    # Infer all element types
    elt_types = [
        _infer_type(elt, type_reg, global_parameter_types, graph, tasks)
        for elt in iterable
    ]

    type_ = types.StructuredType(
        types.TypeStructure(types.StructureType.TUPLE, elt_types)
    )

    return type_


def _infer_type(
    value: Any,
    type_reg: Optional[_TypeMap] = None,
    global_parameter_types: Optional[_TypeMap] = None,
    graph: Optional[_DefMap] = None,
    tasks: Optional[_DefMap] = None,
) -> types.Type:
    """
    Infer the type of the given value.  If the latter four parameters are
    non-None, reference resolution is enabled.  That means if the value is or
    contains strings which look like a reference, they are looked up and
    resolved to types.  If reference resolution is disabled, the string type
    will be inferred for all string values, no matter what they look like.

    Args:
        value: The value to infer a type for
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to
            Type instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short name to task definition

    Returns:
        The value type
    """
    if isinstance(value, str):
        if (
            util.is_reference(value)
            and type_reg is not None
            and global_parameter_types is not None
            and graph is not None
            and tasks is not None
        ):
            type_ = _get_reference_type(
                value[1:], type_reg, global_parameter_types, graph, tasks
            )
        else:
            type_ = type_registry.TYPE_STRING

    elif isinstance(value, bool):
        type_ = type_registry.TYPE_BOOLEAN

    elif isinstance(value, int):
        type_ = type_registry.TYPE_INTEGER

    elif isinstance(value, float):
        type_ = type_registry.TYPE_NUMBER

    elif value is None:
        type_ = type_registry.TYPE_NULL

    elif isinstance(value, Mapping):
        type_ = _infer_type_from_mapping(
            value, type_reg, global_parameter_types, graph, tasks
        )

    elif util.is_iterable(value):
        type_ = _infer_type_from_iterable(
            value, type_reg, global_parameter_types, graph, tasks
        )

    else:
        # If we can't get any more specific, infer the "any" type
        type_ = type_registry.TYPE_ANY

    return type_


def _mapping_structures_compatible(
    invoc_map_struct: types.TypeStructure, task_map_struct: types.TypeStructure
) -> bool:
    """
    Check an invocation mapping structure for compatibility with a task
    parameter mapping structure.

    Args:
        invoc_map_struct: An invocation mapping structure
        task_map_struct: A task parameter mapping structure

    Returns:
        True if the structures are compatible; False if not
    """
    invoc_struct_def = invoc_map_struct.struct_def
    # Whether the mapping structure is of the enumerated property style or the
    # keytype/valuetype style.
    invoc_is_enumerated = isinstance(invoc_struct_def, Mapping)

    task_struct_def = task_map_struct.struct_def
    task_is_enumerated = isinstance(task_struct_def, Mapping)

    if invoc_is_enumerated and task_is_enumerated:
        # Mappings must have the same properties, and property types must
        # be compatible

        # This is redundant because this is what invoc_is_enumerated and
        # task_is_enumerated (being True) means.  The code is more
        # understandable with those variables.  I want readers to not only
        # understand the Python types but the ramifications regarding the type
        # system implemented here: that both the invocation arg and task
        # parameter types are *enumerated* mappings in this case.  But mypy
        # can't see that.
        assert isinstance(invoc_struct_def, Mapping)
        assert isinstance(task_struct_def, Mapping)

        result = len(invoc_struct_def) == len(task_struct_def)

        if result:
            for invoc_prop_name, invoc_prop_type in invoc_struct_def.items():
                task_prop_type = task_struct_def.get(invoc_prop_name)

                if task_prop_type:
                    result = _types_compatible(invoc_prop_type, task_prop_type)
                else:
                    result = False

                if not result:
                    break

    elif invoc_is_enumerated and not task_is_enumerated:
        # Task key type must be string, and all invocation property types
        # must be compatible with the task value type

        # Redundant, for mypy
        assert isinstance(invoc_struct_def, Mapping)
        assert isinstance(task_struct_def, Sequence)

        task_key_type, task_value_type = task_struct_def

        result = task_key_type is type_registry.TYPE_STRING

        if result:
            result = all(
                _types_compatible(invoc_prop_type, task_value_type)
                for invoc_prop_type in invoc_struct_def.values()
            )

    elif not invoc_is_enumerated and task_is_enumerated:
        # Can't pass a keytype/valuetype mapping to an enumerated mapping.

        result = False

    else:
        # Neither mapping structure is enumerated.  Key types and value
        # types must be compatible.

        # Redundant, for mypy
        assert isinstance(invoc_struct_def, Sequence)
        assert isinstance(task_struct_def, Sequence)

        invoc_key_type, invoc_value_type = invoc_struct_def
        task_key_type, task_value_type = task_struct_def

        result = _types_compatible(invoc_key_type, task_key_type) and _types_compatible(
            invoc_value_type, task_value_type
        )

    return result


def _tuple_structures_compatible(
    invoc_tuple_struct: types.TypeStructure, task_tuple_struct: types.TypeStructure
) -> bool:
    """
    Check an invocation tuple structure for compatibility with a task parameter
    tuple structure.

    Args:
        invoc_tuple_struct: An invocation tuple structure
        task_tuple_struct: A task parameter tuple structure

    Returns:
        True if the structures are compatible; False if not
    """
    invoc_struct_def = invoc_tuple_struct.struct_def
    task_struct_def = task_tuple_struct.struct_def

    # For mypy: tuple structures are sequences (the element types)
    assert isinstance(invoc_struct_def, Sequence)
    assert isinstance(task_struct_def, Sequence)

    result = len(invoc_struct_def) == len(task_struct_def)

    if result:
        for invoc_elt_type, task_elt_type in zip(invoc_struct_def, task_struct_def):
            if not _types_compatible(invoc_elt_type, task_elt_type):
                result = False
                break

    return result


def _types_compatible_structured(
    invocation_arg_type: types.StructuredType, task_param_type: types.StructuredType
) -> bool:
    """
    Check an invocation argument type for compatibility with a corresponding
    task parameter type, where both are structured types.

    Args:
        invocation_arg_type: An invocation argument type
        task_param_type: A task parameter argument type

    Returns:
        True if the types are compatible; False if not
    """
    # Special tuple->list cross-type compatibility check.  This is necessary
    # because we will infer a tuple type from a list value, which needs to be
    # able to be compatible with a list type.  Compatibility could be due
    # to inheritance (e.g. tuple[int, number] needs to be compatible with
    # list[number]).
    if (
        invocation_arg_type.structure.struct_type
        is types.StructureType.TUPLE  # noqa: E721
        and task_param_type.structure.struct_type is types.StructureType.LIST
    ):
        tuple_elt_types = invocation_arg_type.structure.struct_def
        list_elt_type = task_param_type.structure.struct_def

        # Redundant, for mypy.
        assert isinstance(tuple_elt_types, Sequence)
        assert isinstance(list_elt_type, types.Type)

        result = all(
            _types_compatible(tuple_elt_type, list_elt_type)
            for tuple_elt_type in tuple_elt_types
        )

    elif (
        invocation_arg_type.structure.struct_type
        is not task_param_type.structure.struct_type
    ):
        # Structure mismatch
        result = False

    # Structure types are the same; must dig into their components and
    # compare
    elif (
        invocation_arg_type.structure.struct_type
        is types.StructureType.MAPPING  # noqa: E721
    ):
        result = _mapping_structures_compatible(
            invocation_arg_type.structure, task_param_type.structure
        )

    elif (
        invocation_arg_type.structure.struct_type
        is types.StructureType.LIST  # noqa: E721
    ):
        # Redundant, for mypy.
        assert isinstance(invocation_arg_type.structure.struct_def, types.Type)
        assert isinstance(task_param_type.structure.struct_def, types.Type)

        result = _types_compatible(
            invocation_arg_type.structure.struct_def,
            task_param_type.structure.struct_def,
        )

    else:
        # Must be tuples
        result = _tuple_structures_compatible(
            invocation_arg_type.structure, task_param_type.structure
        )

    return result


def _types_compatible(
    invocation_arg_type: types.Type, task_param_type: types.Type
) -> bool:
    """
    Check an invocation argument type for compatibility with a corresponding
    task parameter type.

    Args:
        invocation_arg_type: An invocation argument type
        task_param_type: A task parameter argument type

    Returns:
        True if the types are compatible; False if not
    """
    if task_param_type is type_registry.TYPE_ANY:
        # everything is compatible with "any"
        result = True

    elif invocation_arg_type == task_param_type:
        # Simple special case which bypasses the complexity below: if types
        # are the same, they are automatically compatible.

        result = True

    # Union cases

    elif isinstance(invocation_arg_type, types.UnionType):
        # Treat invocation union like an "and" over all union members:
        # all union members must be compatible with the task parameter type.
        result = all(
            _types_compatible(invoc_member_type, task_param_type)
            for invoc_member_type in invocation_arg_type.member_types
        )

    elif isinstance(task_param_type, types.UnionType):
        # Due to the above "if", we know the invocation arg type is not a
        # union here.  It must be compatible with some member of the task
        # parameter union.
        result = any(
            _types_compatible(invocation_arg_type, task_param_member_type)
            for task_param_member_type in task_param_type.member_types
        )

    # Non-union cases

    elif isinstance(invocation_arg_type, types.SimpleType) and isinstance(
        task_param_type, types.SimpleType
    ):
        # Both simple, but not the same (that was checked above).  So we can do
        # a subtype check.
        result = invocation_arg_type.is_subtype_of(task_param_type)

    elif isinstance(invocation_arg_type, types.StructuredType) and isinstance(
        task_param_type, types.StructuredType
    ):
        if invocation_arg_type.name is not None and task_param_type.name is not None:
            # We treat differently named structured types as incompatible,
            # regardless of structural compatibility.  In this case I think we
            # should treat their semantics as different, even if their inner
            # structure aligns.  Names must be unequal here, or they would be
            # the same type, and that case was checked for already.
            result = False

        else:
            result = _types_compatible_structured(invocation_arg_type, task_param_type)

    else:
        # Simple/structured type mismatch
        result = False

    return result


def _check_invocation_parameter(
    invocation_arg_spec: Any,
    task_arg: Mapping[str, Any],
    type_reg: _TypeMap,
    global_parameter_types: _TypeMap,
    graph: _DefMap,
    tasks: _DefMap,
) -> list[ValidationIssue]:
    """
    Check the type of one task invocation parameter from a step against the
    parameter definition declared with the task.

    Args:
        invocation_arg_spec: An argument spec from a task invocation in a step
        task_arg: A task argument definition from a task definition
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to Type
            instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short names to task definitions

    Returns:
        A list of ValidationIssue objects; will be empty if no issues were
        found
    """
    issues = []

    _, task_param_type_name = util.input_def_get_name_type(task_arg)
    # Assume type references have already been validated
    task_param_type = type_reg[task_param_type_name]

    inferred_invocation_arg_type = _infer_type(
        invocation_arg_spec, type_reg, global_parameter_types, graph, tasks
    )

    if not _types_compatible(inferred_invocation_arg_type, task_param_type):
        message = (
            "Inferred invocation arg type and task plugin input type are"
            " incompatible. Found {!s}; expected {!s}"
        ).format(inferred_invocation_arg_type, task_param_type)

        issue = ValidationIssue(IssueType.TYPE, IssueSeverity.ERROR, message)
        issues.append(issue)

    return issues


def _step_check_types(
    step_def: Mapping[str, Any],
    type_reg: _TypeMap,
    global_parameter_types: _TypeMap,
    graph: _DefMap,
    tasks: _DefMap,
) -> list[ValidationIssue]:
    """
    Check the types of task invocation parameters in the given step.

    Args:
        step_def: A step definition
        type_reg: A type registry as a mapping from type name to Type instance
        global_parameter_types: A mapping from global parameter name to Type
            instance, giving the types of all global parameters
        graph: A task graph, mapping step names to step definitions
        tasks: Task definitions, mapping task short name to task definition

    Returns:
        A list of ValidationIssue objects; will be empty if no issues were
        found
    """
    # Get info about how the task plugin is being invoked, in the given step
    (
        invocation_pos_arg_specs,
        invocation_keyword_arg_specs,
    ) = util.step_get_invocation_arg_specs(step_def)

    # For mypy: assume a step definition well-formedness check has already
    # occurred, so we know we can get invocation arguments.
    assert invocation_pos_arg_specs is not None
    assert invocation_keyword_arg_specs is not None

    # Get info about the task plugin: how it needs to be invoked
    task_plugin_short_name = util.step_get_plugin_short_name(step_def)

    # For mypy: assume a step definition well-formedness check has already
    # occurred, so we know we can get a task plugin short name.
    assert task_plugin_short_name is not None

    task_def = tasks[task_plugin_short_name]
    task_inputs_map = util.make_task_input_map(task_def)

    # Now, we can evaluate the task plugin invocation with respect to its
    # parameter type requirements.

    all_issues = []

    # Check invocation positional args
    for arg_idx, (invocation_arg_spec, task_arg) in enumerate(
        zip(
            # rely on correct task_inputs_map iteration order here
            invocation_pos_arg_specs,
            task_inputs_map.values(),
        )
    ):
        issues = _check_invocation_parameter(
            invocation_arg_spec,
            task_arg,
            type_reg,
            global_parameter_types,
            graph,
            tasks,
        )

        # Augment messages with info not known at the point where the message
        # was originally created.
        for issue in issues:
            issue.message = "in positional parameter #{}: {}".format(
                arg_idx + 1, issue.message
            )

        all_issues.extend(issues)

    # Check invocation keyword args
    for (
        invocation_kwarg_name,
        invocation_kwarg_spec,
    ) in invocation_keyword_arg_specs.items():
        # Already checked for invalid keywords in kwarg invocations; may as
        # well not duplicate that check here.
        task_arg = task_inputs_map[invocation_kwarg_name]

        issues = _check_invocation_parameter(
            invocation_kwarg_spec,
            task_arg,
            type_reg,
            global_parameter_types,
            graph,
            tasks,
        )

        for issue in issues:
            issue.message = 'in keyword argument "{}": {}'.format(
                invocation_kwarg_name, issue.message
            )

        all_issues.extend(issues)

    return all_issues


def _check_global_parameter_defaults(
    global_parameter_spec: _DefMap, global_parameter_types: _TypeMap
) -> list[ValidationIssue]:
    """
    Check global parameter default values for compatibility with their
    defined types (if any).

    Args:
        global_parameter_spec: Global parameter specifications taken from an
            experiment description.  This maps a parameter name to a
            definition.
        global_parameter_types: A mapping from global parameter name to
            Type instance, giving the types of all global parameters

    Returns:
        A list of ValidationIssue objects; will be empty if no issues were
            found
    """
    # Can't really use None to mean no default, since that's a valid default!
    no_default = object()

    issues = []

    for (param_name, param_def), param_type in zip(
        # Rely on identical iteration order between the two maps
        global_parameter_spec.items(),
        global_parameter_types.values(),
    ):
        if isinstance(param_def, dict):
            param_default = param_def.get("default", no_default)
        else:
            param_default = param_def

        if param_default is not no_default:
            # Check the default against the type.
            default_type = _infer_type(param_default)

            if not _types_compatible(default_type, param_type):
                message = (
                    'In global parameter "{}": default value {!r} is'
                    " incompatible with its declared type: {!s}"
                ).format(param_name, param_default, param_type)

                issue = ValidationIssue(IssueType.TYPE, IssueSeverity.ERROR, message)
                issues.append(issue)

    return issues


def _infer_global_parameter_types(
    global_parameter_spec: _DefMap, type_reg: _TypeMap
) -> _TypeMap:
    """
    Find types for all global parameters, either by looking up the type name
    if given, or inferring types from default values.

    Args:
        global_parameter_spec: Global parameter specifications taken from an
            experiment description.  This maps a parameter name to a
            definition.
        type_reg: A type registry as a mapping from type name to Type instance

    Returns:
        A mapping from global parameter name to Type object representing the
            types of all global parameters.
    """
    global_parameter_types = {}

    for param_name, param_def in global_parameter_spec.items():
        if isinstance(param_def, dict):
            param_type_name = param_def.get("type")
            param_default = param_def.get("default")

        else:
            param_type_name = None
            param_default = param_def

        if param_type_name:
            param_type = type_reg[param_type_name]

        else:
            # Obtain a type by inference from a default value.  Assume that
            # if a type name is not given, a default value is given, i.e. that
            # check has already occurred.
            param_type = _infer_type(param_default)

        global_parameter_types[param_name] = param_type

    return global_parameter_types


def check_types(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Check the types of all task invocation arguments across all steps in the
    task graph, against declared task plugin inputs and outputs.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be empty if no issues were
        found
    """
    types_ = experiment_desc.get("types", {})
    global_parameter_spec = experiment_desc.get("parameters", {})
    tasks = experiment_desc["tasks"]
    graph = experiment_desc["graph"]

    type_reg = type_registry.build_type_registry(types_)

    global_parameter_types = _infer_global_parameter_types(
        global_parameter_spec, type_reg
    )

    all_issues = _check_global_parameter_defaults(
        global_parameter_spec, global_parameter_types
    )

    if not all_issues:
        for step_name, step_def in graph.items():
            issues = _step_check_types(
                step_def, type_reg, global_parameter_types, graph, tasks
            )

            for issue in issues:
                issue.message = 'in step "{}": {}'.format(step_name, issue.message)

            all_issues.extend(issues)

    return all_issues



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import math
from typing import cast

import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.object_detection.bounding_boxes.coordinates import (
    TensorflowBoundingBoxesBatchedGrid,
)
from dioptra.sdk.object_detection.bounding_boxes.postprocessing.bounding_boxes_postprocessing import (  # noqa: B950
    BoundingBoxesYOLOV1PostProcessing,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import Tensor
    from tensorflow.image import combined_non_max_suppression

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowBoundingBoxesYOLOV1NMS(BoundingBoxesYOLOV1PostProcessing):
    def __init__(
        self,
        bounding_boxes_batched_grid: TensorflowBoundingBoxesBatchedGrid,
        max_output_size_per_class: int,
        max_total_size: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> None:
        self._bbox_batched_grid = bounding_boxes_batched_grid
        self._max_output_size_per_class = max_output_size_per_class
        self._max_total_size = max_total_size
        self._iou_threshold = iou_threshold
        self._score_threshold = score_threshold

    @classmethod
    def on_grid_shape(
        cls,
        grid_shape: tuple[int, int],
        max_output_size_per_class: int = 20,
        iou_threshold: float = 0.5,
        score_threshold: float = 0.5,
    ) -> TensorflowBoundingBoxesYOLOV1NMS:
        return cls(
            bounding_boxes_batched_grid=(
                TensorflowBoundingBoxesBatchedGrid.on_grid_shape(grid_shape=grid_shape)
            ),
            max_output_size_per_class=max_output_size_per_class,
            max_total_size=math.prod(grid_shape),
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def postprocess(
        self, bboxes_cell_xywh: Tensor, bboxes_conf: Tensor, bboxes_labels: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        batch_size = tf.cast(tf.shape(bboxes_cell_xywh)[0], tf.int32)
        num_boxes = tf.cast(tf.reduce_prod(tf.shape(bboxes_cell_xywh)[1:4]), tf.int32)
        num_labels = tf.cast(tf.shape(bboxes_labels)[-1], tf.int32)

        boxes = tf.reshape(
            self._from_cell_xywh_to_corner(bboxes_cell_xywh=bboxes_cell_xywh),
            shape=(batch_size, num_boxes, 1, 4),
        )
        scores = tf.reshape(
            self._calculate_prediction_scores(
                bboxes_conf=bboxes_conf, bboxes_labels=bboxes_labels
            ),
            shape=(batch_size, num_boxes, num_labels),
        )

        return cast(
            tuple[Tensor, Tensor, Tensor, Tensor],
            combined_non_max_suppression(
                boxes=boxes,
                scores=scores,
                max_output_size_per_class=self._max_output_size_per_class,
                max_total_size=self._max_total_size,
                iou_threshold=self._iou_threshold,
                score_threshold=self._score_threshold,
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.int32),
            tf.TensorSpec(None, tf.int32),
        ]
    )
    def embed(
        self, bboxes_corner: Tensor, bboxes_labels: Tensor, n_classes: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        return cast(
            tuple[Tensor, Tensor, Tensor, Tensor],
            self._bbox_batched_grid.embed(
                bboxes_corner=bboxes_corner,
                bboxes_labels=bboxes_labels,
                n_classes=n_classes,
            ),
        )

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _from_cell_xywh_to_corner(self, bboxes_cell_xywh: Tensor) -> Tensor:
        n_bounding_boxes = tf.cast(tf.shape(bboxes_cell_xywh)[-2], tf.int32)
        bboxes_corner = self._bbox_batched_grid.from_cell_xywh_to_corner(
            bboxes_cell_xywh=bboxes_cell_xywh,
            n_bounding_boxes=n_bounding_boxes,
        )

        return tf.gather(bboxes_corner, [1, 0, 3, 2], axis=-1)

    @tf.function(
        input_signature=[
            tf.TensorSpec(None, tf.float32),
            tf.TensorSpec(None, tf.float32),
        ]
    )
    def _calculate_prediction_scores(
        self, bboxes_conf: Tensor, bboxes_labels: Tensor
    ) -> Tensor:
        n_bounding_boxes = tf.cast(tf.shape(bboxes_conf)[-1], tf.int32)
        bboxes_labels = tf.tile(
            tf.expand_dims(bboxes_labels, axis=-2),
            multiples=(1, 1, 1, n_bounding_boxes, 1),
        )

        return tf.expand_dims(bboxes_conf, axis=-1) * bboxes_labels



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
#
# This is a fork of the work
# https://github.com/gahjelle/pyplugs/blob/90e635777672f75080291c737f08453a26ea380d/tests/plugin_directory/__init__.py  # noqa: B950
# See copyright below.
#
# Copyright (c) 2019 Geir Arne Hjelle
# Distributed under the terms of the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Sample plugins used by the dioptra.pyplugs tests"""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import pytest

from dioptra.task_engine.issues import IssueSeverity
from dioptra.task_engine.validation import is_valid, validate


@pytest.mark.parametrize(
    "experiment_desc",
    [
        None,
        1,
        True,
        "foo",
        {},
        [],
        {"graph": {}},
        {"tasks": {}},
        {"foo": "bar"},
        ["foo"],
        {
            "tasks": {"add": {"plugin": "org.example.add"}},
            "graph": {"step1": {"add": []}},
            "extra": "property",
        },
    ],
)
def test_invalid_description(experiment_desc) -> None:
    # I don't test the error messages; they are just prose-y strings, and
    # we should be able to change wording without breaking a bunch of unit
    # tests, right?
    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "parameters",
    [
        {},
        {"arg1": None},
        {"arg1": "default_value"},
        {"arg1": {"default": None}},
        {"arg1": {"default": "default_value"}},
        {"arg1": {"type": "integer"}},
        {"arg1": {"default": "default_value", "type": "string"}},
    ],
)
def test_valid_parameters(parameters) -> None:
    # Since we are strictly testing the parameters part, we just parameterize
    # that part, and insert it into an otherwise valid description.
    experiment_desc = {
        "parameters": parameters,
        "tasks": {
            "add": {
                "plugin": "org.example.add",
                "inputs": [
                    {"n1": "number"},
                    {"n2": "number"},
                    {"n3": "number"},
                ],
            }
        },
        "graph": {"step1": {"add": [1, 2, 3]}},
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "parameters",
    [
        None,
        1,
        [1],
        "foo",
        {"foo.bar": 1},
        {"arg1": {"foo": 1}},
        {1: {"default": "foo"}},
        {"foo": {"type": None}},
        {"foo": {"type": "badtype", "default": "value"}},
        {"foo": {"type": "integer", "default": "value"}},
        {"step1": "integer"},  # name collides with a graph step
    ],
)
def test_invalid_parameters(parameters) -> None:
    experiment_desc = {
        "parameters": parameters,
        "tasks": {"add": {"plugin": "org.example.add"}},
        "graph": {"step1": {"add": []}},
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "tasks",
    [
        {
            "add": {
                "plugin": "org.example.add",
                "inputs": [
                    {"n1": "integer"},
                    {"n2": "integer"},
                    {"n3": "integer"},
                ],
            }
        },
        {
            "add": {
                "plugin": "org.example.add",
                "inputs": [
                    {"n1": "integer"},
                    {"n2": "integer"},
                    {"n3": "integer"},
                ],
                "outputs": {"foo": "string"},
            }
        },
        {
            "add": {
                "plugin": "org.example.add",
                "inputs": [
                    {"n1": "integer"},
                    {"n2": "integer"},
                    {"name": "n3", "type": "number", "required": True},
                ],
                "outputs": [{"foo": "string"}, {"bar": "integer"}],
            }
        },
        {
            "add": {
                "plugin": "org.example.add",
                "inputs": [
                    {"n1": "integer"},
                    {"n2": "integer"},
                    {"n3": "integer"},
                ],
            },
            "sub": {"plugin": "org.example.sub"},
        },
    ],
)
def test_valid_tasks(tasks) -> None:
    experiment_desc = {"tasks": tasks, "graph": {"step1": {"add": [1, 2, 3]}}}

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "tasks",
    [
        None,
        1,
        "foo",
        [],
        {},
        {"add": 1},
        {1: "add"},
        {1: {"plugin": "org.example.add"}},
        {"add": []},
        {"add": {}},
        {"add": {"foo": True}},
        {"add": {True: "foo"}},
        {"add": {"plugin": ""}},
        {"add": {"plugin": "abc"}},
        {"add": {"plugin": "org.example.add."}},
        {"add": {"plugin": "org.example.add", "foo": 123}},
        {"add": {"plugin": "org.example.add", "outputs": {"value": "badtype"}}},
        {"add": {"plugin": "org.example.add", "inputs": [{"value": "badtype"}]}},
        {
            "add": {
                "plugin": "org.example.add",
                "outputs": [{"value": "number"}, {"value": "number"}],
            }
        },
        {
            "add": {
                "plugin": "org.example.add",
                "inputs": [{"value": "number"}, {"value": "number"}],
            }
        },
    ],
)
def test_invalid_tasks(tasks) -> None:
    experiment_desc = {"tasks": tasks, "graph": {"step1": {"add": []}}}

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"0in": []}},
        {"step1": {"0in": {}}},
        {"step1": {"1in": 1}},
        {"step1": {"1in": [1]}},
        {"step1": {"1in": {"in1": 1}}},
        {"step1": {"2in": [1, 2]}},
        {"step1": {"2in": {"in1": 1, "in2": 2}}},
        {"step1": {"task": "0in"}},
        {"step1": {"task": "1in", "args": 1}},
        {"step1": {"task": "1in", "args": [1]}},
        {"step1": {"task": "2in", "args": [1, 2]}},
        {"step1": {"task": "2in", "kwargs": {"in1": 1, "in2": 2}}},
        {"step1": {"task": "2in", "args": 1, "kwargs": {"in2": 2}}},
        {"step1": {"1in": [1]}, "step2": {"2in": [3, 4]}},
    ],
)
def test_valid_graph(graph) -> None:
    experiment_desc = {
        "tasks": {
            "0in": {"plugin": "org.example.task0in"},
            "1in": {"plugin": "org.example.task1in", "inputs": [{"in1": "number"}]},
            "2in": {
                "plugin": "org.example.task1in",
                "inputs": [
                    {"in1": "number"},
                    {"in2": "number"},
                ],
            },
        },
        "graph": graph,
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        None,
        1,
        "foo",
        [],
        {},
        {"step1": 1},
        {1: "step1"},
        {1: {"add": []}},
        {"step1": [1]},
        {"step1": {"sub": 1}},
        {"step1": {"task": "sub"}},
        {"step1": {"add": [], "2arg": [1, "foo"]}},
        {
            "step1": {
                # missing task short name
                "dependencies": ["step2"]
            },
            "step2": {"add": 1},
        },
        {"step1": {"task": "add", "foo": "bar"}},
        {
            "step1": {
                "task": "2arg",
                "args": [1, "foo"],
                "kwargs": {
                    # arg1 set both positionally and via keyword
                    "arg1": 1
                },
            }
        },
        {
            "step1": {
                "task": "2arg",
                "args": [1, "foo"],
                "kwargs": {
                    # arg1 and arg2 set both positionally and via keyword
                    "arg1": 1,
                    "arg2": "foo",
                },
            }
        },
    ],
)
def test_invalid_graph(graph) -> None:
    experiment_desc = {
        "tasks": {
            "add": {"plugin": "org.example.add"},
            "2arg": {
                "plugin": "org.example.task2arg",
                "inputs": [{"arg1": "number"}, {"arg2": "string"}],
            },
        },
        "graph": graph,
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"1out": []}, "step2": {"1in": "$step1"}},
        {
            "step1": {"1out": []},
            "step2": {"2in": ["$step1", "$step3.out2"]},
            "step3": {"2out": []},
        },
        {
            "step1": {"1out": []},
            "step2": {"1out": []},
            "step3": {"2inmap": ["$step1", {"arg1": "$step2"}]},
        },
        {
            "step1": {"1out": []},
            "step2": {
                "2inmaplist": {"in1": "$step1", "in2": {"subarg1": ["$step3.out2"]}}
            },
            "step3": {"2out": []},
        },
        {
            "step1": {"1out": []},
            "step2": {
                "task": "2in",
                "args": ["$step1.out"],
                "kwargs": {"in2": "$step3.out1"},
            },
            "step3": {"2out": []},
        },
    ],
)
def test_valid_step_references(graph) -> None:
    experiment_desc = {
        "types": {
            "str_to_num": {"mapping": ["string", "number"]},
            "str_to_list_num": {"mapping": ["string", {"list": "number"}]},
        },
        "tasks": {
            "1out": {"plugin": "org.example.task1out", "outputs": {"out": "number"}},
            "2out": {
                "plugin": "org.example.task2out",
                "outputs": [{"out1": "number"}, {"out2": "number"}],
            },
            "1in": {"plugin": "org.example.task1in", "inputs": [{"in": "number"}]},
            "2in": {
                "plugin": "org.example.task2in",
                "inputs": [{"in1": "number"}, {"in2": "number"}],
            },
            "2inmap": {
                "plugin": "org.example.task2inmap",
                "inputs": [{"in1": "number"}, {"in2": "str_to_num"}],
            },
            "2inmaplist": {
                "plugin": "org.example.task2inmaplist",
                "inputs": [{"in1": "number"}, {"in2": "str_to_list_num"}],
            },
        },
        "graph": graph,
    }

    assert is_valid(experiment_desc)


def test_output_named_name():
    # "name" has special meaning in an input (it's a long form input);
    # ensure it does not have that meaning in an output and I didn't mess
    # it up anywhere...
    experiment_desc = {
        "tasks": {
            "nameout": {"plugin": "org.example.nameout", "outputs": {"name": "string"}},
            "nameoutarr": {
                "plugin": "org.example.nameout",
                "outputs": [{"name": "string"}],
            },
            "1in": {
                "plugin": "org.example.1in",
                "inputs": [{"name": "in", "type": "string"}],
            },
        },
        "graph": {
            "singleout": {"nameout": []},
            "arrayout": {"nameoutarr": []},
            "in_1": {"1in": "$singleout"},
            "in_2": {"1in": "$arrayout"},
            "in_3": {"1in": "$singleout.name"},
            "in_4": {"1in": "$arrayout.name"},
        },
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"1innum": "$param1"}},
        {"step1": {"1innestedthing": {"foo": [{"bar": ["$param2"]}]}}},
    ],
)
def test_valid_parameter_references(graph) -> None:
    experiment_desc = {
        "parameters": {
            "param1": 1.23,  # inferred type: number
            "param2": {"type": "string"},
        },
        "types": {"nestedthing": {"tuple": [{"mapping": {"bar": {"list": "string"}}}]}},
        "tasks": {
            "1innum": {"plugin": "org.example.task1", "inputs": [{"in": "number"}]},
            "1innestedthing": {
                "plugin": "org.example.task2",
                "inputs": [{"foo": "nestedthing"}],
            },
        },
        "graph": graph,
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"task1": "$step2"}},
        {"step1": {"task1": 1}, "step2": {"task2": "$step1.foo"}},
        {"step1": {"task1": {"arg1": "$step3"}}, "step3": {"task3": []}},
        {"step1": {"task1": "$param3"}},
        {"step1": {"task1": "$param1.value"}},
        {"step1": {"task4": []}, "step2": {"task1": "$step1"}},
        {"step1": {"task4": []}, "step2": {"task1": "$step1.value1"}},
    ],
)
def test_invalid_references(graph) -> None:
    experiment_desc = {
        "parameters": {"param1": "number", "param2": "number"},
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "inputs": [{"arg1": "number"}],
                "outputs": {"value": "number"},
            },
            "task2": {
                "plugin": "org.example.task2",
                "inputs": [{"in": "number"}],
                "outputs": {"value": "number"},
            },
            "task3": {
                "plugin": "org.example.task3",
                "outputs": [{"value1": "number"}, {"value2": "number"}],
            },
            "task4": {"plugin": "org.example.task4"},
        },
        "graph": graph,
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"task1": "$step1"}},
        {
            "step1": {"task1": ["$step2"]},
            "step2": {"task2": {"foo": "$step1.value"}},
        },
        {
            "step1": {"task": "task1", "args": ["$step3.value2"]},
            "step2": {"task2": {"foo": "$step1"}},
            "step3": {"task3": {"bar": {"baz": ["$step2.value"]}}},
        },
        {"step1": {"task1": 1, "dependencies": ["step1"]}},
        {
            "step1": {"task1": 1, "dependencies": ["step3"]},
            "step2": {"task2": 2, "dependencies": ["step1"]},
            "step3": {"task3": [{"abc": [1]}], "dependencies": ["step2"]},
        },
    ],
)
def test_step_cycle(graph) -> None:
    experiment_desc = {
        "types": {"str_to_list_num": {"mapping": ["string", {"list": "number"}]}},
        "tasks": {
            "task1": {
                "plugin": "org.example.task1",
                "inputs": [{"in": "number"}],
                "outputs": {"value": "number"},
            },
            "task2": {
                "plugin": "org.example.task2",
                "inputs": [{"foo": "number"}],
                "outputs": [{"value": "number"}],
            },
            "task3": {
                "plugin": "org.example.task3",
                "inputs": [{"bar": "str_to_list_num"}],
                "outputs": [{"value1": "number"}, {"value2": "number"}],
            },
        },
        "graph": graph,
    }

    assert not is_valid(experiment_desc)


def test_name_collision() -> None:
    # parameter name collides with a step name
    experiment_desc = {
        "parameters": {"foo": 1},
        "tasks": {"task1": {"plugin": "org.example.foo", "inputs": [{"in": "number"}]}},
        "graph": {"foo": {"task1": 1}},
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"task1": [], "dependencies": ["step2"]}, "step2": {"task1": []}},
        {
            "step1": {"task1": [], "dependencies": ["step2", "step3"]},
            "step2": {"task1": [], "dependencies": ["step3"]},
            "step3": {"task1": []},
        },
        {"step1": {"task1": [], "dependencies": "step2"}, "step2": {"task1": []}},
    ],
)
def test_valid_dependencies(graph) -> None:
    experiment_desc = {
        "tasks": {"task1": {"plugin": "org.example.foo"}},
        "graph": graph,
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"task1": [], "dependencies": ["step2"]}},
        {
            "step1": {"task1": [], "dependencies": ["step2", "step3"]},
            "step2": {"task1": []},
        },
        {"step1": {"task1": [], "dependencies": 1}},
    ],
)
def test_invalid_dependencies(graph) -> None:
    experiment_desc = {
        "tasks": {"task1": {"plugin": "org.example.foo"}},
        "graph": graph,
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {"allopt": {}}},
        {"step1": {"allopt": 1}},
        {"step1": {"allopt": {"in2": 2}}},
        {"step1": {"allopt": [1, 2]}},
        {"step1": {"allopt": {"in1": 1, "in2": 2}}},
        {"step1": {"1stopt": {"in2": 2}}},
        {"step1": {"task": "1stopt", "kwargs": {"in2": 2}}},
        {"step1": {"1stopt": [1, 2]}},
        {"step1": {"2ndopt": 1}},
        {"step1": {"2ndopt": [1]}},
        {"step1": {"2ndopt": {"in1": 1}}},
        {"step1": {"task": "2ndopt", "args": 1, "kwargs": {"in2": 2}}},
    ],
)
def test_optional_task_inputs_valid(graph) -> None:
    experiment_desc = {
        "tasks": {
            "allopt": {
                "plugin": "org.example.taskallopt",
                "inputs": [
                    {"name": "in1", "type": "number", "required": False},
                    {"name": "in2", "type": "number", "required": False},
                ],
            },
            "1stopt": {
                "plugin": "org.example.task1stopt",
                "inputs": [
                    {"name": "in1", "type": "number", "required": False},
                    {"name": "in2", "type": "number", "required": True},
                ],
            },
            "2ndopt": {
                "plugin": "org.example.task2ndopt",
                "inputs": [
                    {"name": "in1", "type": "number", "required": True},
                    {"name": "in2", "type": "number", "required": False},
                ],
            },
        },
        "graph": graph,
    }

    assert is_valid(experiment_desc)


@pytest.mark.parametrize(
    "graph",
    [
        {"step1": {}},
        {"step1": {"1stopt": 1}},
        {"step1": {"1stopt": {"in1": 1}}},
        {"step1": {"task": "1stopt", "args": [1]}},
        {"step1": {"task": "1stopt", "kwargs": {"in1": 1}}},
        {"step1": {"2ndopt": {"in2": 1}}},
        {"step1": {"task": "2ndopt", "kwargs": {"in2": 1}}},
    ],
)
def test_optional_task_inputs_invalid(graph) -> None:
    experiment_desc = {
        "tasks": {
            "1stopt": {
                "plugin": "org.example.task1stopt",
                "inputs": [
                    {"name": "in1", "type": "number", "required": False},
                    {"name": "in2", "type": "number", "required": True},
                ],
            },
            "2ndopt": {
                "plugin": "org.example.task2ndopt",
                "inputs": [
                    {"name": "in1", "type": "number", "required": True},
                    {"name": "in2", "type": "number", "required": False},
                ],
            },
        },
        "graph": graph,
    }

    assert not is_valid(experiment_desc)


def test_invocation_too_many_positional_args() -> None:
    experiment_desc = {
        "tasks": {
            "atask": {"plugin": "org.example.task", "inputs": [{"in1": "number"}]}
        },
        "graph": {"step1": {"atask": [1, 2, 3]}},
    }

    assert not is_valid(experiment_desc)


def test_invocation_bad_kwarg() -> None:
    experiment_desc = {
        "tasks": {
            "atask": {"plugin": "org.example.task", "inputs": [{"in1": "number"}]}
        },
        "graph": {"step1": {"atask": {"badkwarg": 1}}},
    }

    assert not is_valid(experiment_desc)


@pytest.mark.parametrize(
    "types",
    [
        {"A": {"is_a": "A"}},
        {"A": {"list": "A"}},
        {"A": {"tuple": ["A"]}},
        {"A": {"mapping": ["string", "A"]}},
        {"A": {"union": ["A"]}},
        {"A": {"is_a": "B"}, "B": {"is_a": "A"}},
        {
            "A": {"list": "B"},
            "B": {"mapping": ["string", {"tuple": ["number", "C"]}]},
            "C": {"union": ["string", "boolean", "A"]},
        },
    ],
)
def test_type_reference_cycle(types) -> None:
    experiment_desc = {
        "types": types,
        "tasks": {"foo": {"plugin": "org.example.foo"}},
        "graph": {"step1": {"foo": []}},
    }

    assert not is_valid(experiment_desc)


def test_union_dupes():
    experiment_desc = {
        "types": {
            "A": None,
            "B": None,
            "C": None,
            "dupe_union": {"union": ["A", "B", "C", "C", "A"]},
        },
        "tasks": {"task1": {"plugin": "org.example.task1"}},
        "graph": {"step1": {"task1": []}},
    }

    issues = validate(experiment_desc)

    # Should only get a warning for the dupe_union type.
    assert len(issues) == 1 and issues[0].severity is IssueSeverity.WARNING

    # Since this is only a warning, ensure that is_valid() returns True, even
    # though there are issues.
    assert is_valid(experiment_desc)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import io
import tarfile
import timeit
from pathlib import Path
from typing import List, Mapping, Optional, Union

PathLike = List[Union[str, Path]]


class Timer(object):
    def __init__(self, timeout: Optional[float] = None) -> None:
        self.start = timeit.default_timer()
        self._elapsed: Optional[float] = None
        self._timeout = timeout

    def __enter__(self) -> Timer:
        self.start = timeit.default_timer()
        self._elapsed = None
        return self

    def __exit__(self, *args) -> None:
        self._elapsed = timeit.default_timer() - self.start

    @property
    def elapsed(self) -> float:
        if self._elapsed is not None:
            return self._elapsed

        return timeit.default_timer() - self.start

    @property
    def timeout_exceeded(self) -> bool:
        if self._timeout is None:
            return False

        return self._timeout < self.elapsed


def make_tarball_bytes(tar_content: Mapping[str, bytes]) -> bytes:
    """
    Make a tarball containing the given content.  Return the tarball content
    as bytes.

    Args:
        tar_content: A mapping from file path to bytes, which lays out the
            paths and file content which should go into the tarball.

    Returns:
        Tarball bytes
    """

    tar_content_stream = io.BytesIO()
    tf = tarfile.TarFile.open(fileobj=tar_content_stream, mode="w:gz")

    for filepath, file_content in tar_content.items():
        ti = tarfile.TarInfo(filepath)
        ti.size = len(file_content)

        tf.addfile(ti, io.BytesIO(file_content))

    tf.close()

    return tar_content_stream.getvalue()



"""empty message

Revision ID: a278f46a1c9d
Revises: 6a343ff9bc2f
Create Date: 2020-09-29 20:09:23.101994

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import column, table

# revision identifiers, used by Alembic.
revision = "a278f46a1c9d"
down_revision = "6a343ff9bc2f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.add_column(sa.Column("is_deleted", sa.Boolean(), nullable=True))

    role = table("experiments", column("is_deleted"))
    op.execute(role.update().values(is_deleted=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.drop_column("is_deleted")

    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Common schemas for serializing/deserializing resources."""

from marshmallow import Schema, fields

from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    DraftTypeQueryParametersSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
)


class DraftRefSchema(Schema):
    """The reference schema for the data stored in a Draft."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the Draft."),
        dump_only=True,
    )
    resourceType = fields.String(
        attribute="resource_type",
        metadata=dict(description="The resource type of the Draft."),
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full Draft."),
        relative=True,
    )


class DraftSchema(Schema):
    """A base schema for a draft of a resource."""

    from dioptra.restapi.v1.groups.schema import GroupRefSchema
    from dioptra.restapi.v1.users.schema import UserRefSchema

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID of the Draft."),
        dump_only=True,
    )
    group = fields.Nested(
        GroupRefSchema,
        attribute="group",
        metadata=dict(description="Group that owns the draft resource."),
        dump_only=True,
    )
    user = fields.Nested(
        UserRefSchema,
        attribute="user",
        metadata=dict(description="User that created the draft resource."),
        dump_only=True,
    )
    payload = fields.Dict(
        attribute="payload",
        metadata=dict(description="The contents of the draft resource."),
        dump_only=True,
    )
    createdOn = fields.DateTime(
        attribute="created_on",
        metadata=dict(description="Timestamp when the draft resource was created."),
        dump_only=True,
    )
    lastModifiedOn = fields.DateTime(
        attribute="last_modified_on",
        metadata=dict(
            description="Timestamp when the draft resource was last modified."
        ),
        dump_only=True,
    )
    resourceType = fields.String(
        attribute="resource_type",
        metadata=dict(description="The type of resource of this draft."),
        dump_only=True,
    )
    resource = fields.Integer(
        attribute="resource_id",
        metadata=dict(description="ID of the resource this draft modifies."),
        dump_only=True,
    )
    resourceSnapshot = fields.Integer(
        attribute="resource_snapshot_id",
        metadata=dict(description="ID of the resource snapshot this draft modifies."),
        dump_only=True,
        allow_none=True,
    )
    metadata = fields.Dict(
        attribute="metadata",
        metadata=dict(description="Additional metadata about the draft"),
        dump_only=True,
    )


class DraftPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Draft."""

    data = fields.Nested(
        DraftSchema,
        many=True,
        metadata=dict(description="List of Drafts in the current page."),
    )


class DraftGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    DraftTypeQueryParametersSchema,
):
    """The query parameters for the GET method of the /<resource>/drafts endpoint."""
